@@ -30,6 +30,7 @@
 
 static PyObject *gerror_exc = NULL;
 static gboolean use_gil_state_api = FALSE;
+static PyObject *_pyg_signal_accumulator_true_handled_func;
 
 GQuark pyginterface_type_key;
 GQuark pygobject_class_init_key;
@@ -458,6 +459,61 @@ pyg_object_class_init(GObjectClass *class, PyObject *py_class)
     class->get_property = pyg_object_get_property;
 }
 
+typedef struct _PyGSignalAccumulatorData {
+    PyObject *callable;
+    PyObject *user_data;
+} PyGSignalAccumulatorData;
+
+static gboolean
+_pyg_signal_accumulator(GSignalInvocationHint *ihint,
+                        GValue *return_accu,
+                        const GValue *handler_return,
+                        gpointer _data)
+{
+    PyObject *py_ihint, *py_return_accu, *py_handler_return, *py_detail;
+    PyObject *py_retval;
+    gboolean retval = FALSE;
+    PyGSignalAccumulatorData *data = _data;
+    PyGILState_STATE state;
+
+    state = pyg_gil_state_ensure();
+    if (ihint->detail)
+        py_detail = PyString_FromString(g_quark_to_string(ihint->detail));
+    else {
+        Py_INCREF(Py_None);
+        py_detail = Py_None;
+    }
+
+    py_ihint = Py_BuildValue("lNi", (long int) ihint->signal_id,
+                             py_detail, ihint->run_type);
+    py_handler_return = pyg_value_as_pyobject(handler_return, TRUE);
+    py_return_accu = pyg_value_as_pyobject(return_accu, FALSE);
+    if (data->user_data)
+        py_retval = PyObject_CallFunction(data->callable, "NNNO", py_ihint,
+                                          py_return_accu, py_handler_return,
+                                          data->user_data);
+    else
+        py_retval = PyObject_CallFunction(data->callable, "NNN", py_ihint,
+                                          py_return_accu, py_handler_return);
+    if (!py_retval)
+	PyErr_Print();
+    else {
+        if (!PyTuple_Check(py_retval) || PyTuple_Size(py_retval) != 2) {
+            PyErr_SetString(PyExc_TypeError, "accumulator function must return"
+                            " a (bool, object) tuple");
+            PyErr_Print();
+        } else {
+            retval = PyObject_IsTrue(PyTuple_GET_ITEM(py_retval, 0));
+            if (pyg_value_from_pyobject(return_accu, PyTuple_GET_ITEM(py_retval, 1))) {
+                PyErr_Print();
+            }
+        }
+        Py_DECREF(py_retval);
+    }
+    pyg_gil_state_release(state);
+    return retval;
+}
+
 static gboolean
 create_signal (GType instance_type, const gchar *signal_name, PyObject *tuple)
 {
@@ -467,9 +523,13 @@ create_signal (GType instance_type, const gchar *signal_name, PyObject *tuple)
     guint n_params, i;
     GType *param_types;
     guint signal_id;
+    GSignalAccumulator accumulator = NULL;
+    PyGSignalAccumulatorData *accum_data = NULL;
+    PyObject *py_accum = NULL, *py_accum_data = NULL;
 
-    if (!PyArg_ParseTuple(tuple, "iOO", &signal_flags, &py_return_type,
-			  &py_param_types)) {
+    if (!PyArg_ParseTuple(tuple, "iOO|OO", &signal_flags, &py_return_type,
+			  &py_param_types, &py_accum, &py_accum_data))
+    {
 	gchar buf[128];
 
 	PyErr_Clear();
@@ -478,6 +538,15 @@ create_signal (GType instance_type, const gchar *signal_name, PyObject *tuple)
 	return FALSE;
     }
 
+    if (py_accum && py_accum != Py_None && !PyCallable_Check(py_accum))
+    {
+	gchar buf[128];
+
+	g_snprintf(buf, sizeof(buf), "accumulator for __gsignals__['%s'] must be callable", signal_name);
+	PyErr_SetString(PyExc_TypeError, buf);
+	return FALSE;
+    }
+
     return_type = pyg_type_from_object(py_return_type);
     if (!return_type)
 	return FALSE;
@@ -502,9 +571,22 @@ create_signal (GType instance_type, const gchar *signal_name, PyObject *tuple)
 	Py_DECREF(item);
     }
 
+    if (py_accum == _pyg_signal_accumulator_true_handled_func)
+        accumulator = g_signal_accumulator_true_handled;
+    else {
+        if (py_accum != NULL && py_accum != Py_None) {
+            accum_data = g_new(PyGSignalAccumulatorData, 1);
+            accum_data->callable = py_accum;
+            Py_INCREF(py_accum);
+            accum_data->user_data = py_accum_data;
+            Py_XINCREF(py_accum_data);
+            accumulator = _pyg_signal_accumulator;
+        }
+    }
+
     signal_id = g_signal_newv(signal_name, instance_type, signal_flags,
 			      pyg_signal_class_closure_get(),
-			      (GSignalAccumulator)0, NULL,
+			      accumulator, accum_data,
 			      (GSignalCMarshaller)0,
 			      return_type, n_params, param_types);
     g_free(param_types);
@@ -695,7 +777,7 @@ create_property (const gchar  *prop_name,
 	{
 	    guint64 minimum, maximum, default_value;
 
-	    if (!PyArg_ParseTuple(args, "LLL", &minimum, &maximum,
+	    if (!PyArg_ParseTuple(args, "KKK", &minimum, &maximum,
 				  &default_value))
 		return NULL;
 	    pspec = g_param_spec_uint64 (prop_name, nick, blurb, minimum,
@@ -1639,8 +1721,8 @@ pyg_object_new (PyGObject *self, PyObject *args, PyObject *kwargs)
 	    if (pyg_param_gvalue_from_pyobject(&params[n_params].value,
 					       value, pspec) < 0) {
 		PyErr_Format(PyExc_TypeError,
-			     "could not convert value for property `%s'",
-			     key_str);
+			     "could not convert value for property `%s' from %s to %s",
+			     key_str, value->ob_type->tp_name, g_type_name(G_PARAM_SPEC_VALUE_TYPE(pspec)));
 		goto cleanup;
 	    }
 	    params[n_params].name = g_strdup(key_str);
@@ -1839,6 +1921,12 @@ iowatch_marshal(GIOChannel *source, GIOCondition condition, gpointer user_data)
 	PyErr_Print();
 	res = FALSE;
     } else {
+        if (ret == Py_None) {
+            if (PyErr_Warn(PyExc_Warning, "gobject.io_add_watch callback returned None;"
+                           " should return True/False")) {
+                PyErr_Print();
+            }
+        }
 	res = PyObject_IsTrue(ret);
 	Py_DECREF(ret);
     }
@@ -1889,7 +1977,7 @@ pyg_io_add_watch(PyObject *self, PyObject *args, PyObject *kwargs)
     iochannel = g_io_channel_unix_new(fd);
     handler_id = g_io_add_watch_full(iochannel, priority, condition,
 				     iowatch_marshal, data,
-				    (GDestroyNotify)pyg_destroy_notify);
+				     (GDestroyNotify)pyg_destroy_notify);
     g_io_channel_unref(iochannel);
     
     return PyInt_FromLong(handler_id);
@@ -2231,6 +2319,149 @@ pyg_main_depth(PyObject *unused)
     return PyInt_FromLong(g_main_depth());
 }
 
+static PyObject *
+pyg_signal_accumulator_true_handled(PyObject *unused, PyObject *args)
+{
+    PyErr_SetString(PyExc_TypeError, "signal_accumulator_true_handled can only"
+                    " be used as accumulator argument when registering signals");
+    return NULL;
+}
+
+static gboolean
+marshal_emission_hook(GSignalInvocationHint *ihint,
+		      guint n_param_values,
+		      const GValue *param_values,
+		      gpointer user_data)
+{
+    PyGILState_STATE state;
+    gboolean retval = FALSE;
+    PyObject *func, *args;
+    PyObject *retobj;
+    PyObject *params;
+    guint i;
+
+    state = pyg_gil_state_ensure();
+
+    /* construct Python tuple for the parameter values */
+    params = PyTuple_New(n_param_values);
+
+    for (i = 0; i < n_param_values; i++) {
+	PyObject *item = pyg_value_as_pyobject(&param_values[i], FALSE);
+	
+	/* error condition */
+	if (!item) {
+	    goto out;
+	}
+	PyTuple_SetItem(params, i, item);
+    }
+
+    args = (PyObject *)user_data;
+    func = PyTuple_GetItem(args, 0);
+    args = PySequence_Concat(params, PyTuple_GetItem(args, 1));
+    Py_DECREF(params);
+
+    /* params passed to function may have extra arguments */
+
+    retobj = PyObject_CallObject(func, args);
+    Py_DECREF(args);
+    if (retobj == NULL) {
+        PyErr_Print();
+    }
+    
+    retval = (retobj == Py_True ? TRUE : FALSE);
+    Py_XDECREF(retobj);
+out:
+    pyg_gil_state_release(state);
+    return retval;
+}
+
+static PyObject *
+pyg_add_emission_hook(PyGObject *self, PyObject *args)
+{
+    PyObject *first, *callback, *extra_args, *data;
+    gchar *name;
+    gulong hook_id;
+    guint sigid, len;
+    GQuark detail = 0;
+    GType gtype;
+    PyObject *pygtype;
+
+    len = PyTuple_Size(args);
+    if (len < 3) {
+	PyErr_SetString(PyExc_TypeError,
+			"kiwi.add_emission_hook requires at least 3 arguments");
+	return NULL;
+    }
+    first = PySequence_GetSlice(args, 0, 3);
+    if (!PyArg_ParseTuple(first, "OsO:add_emission_hook",
+			  &pygtype, &name, &callback)) {
+	Py_DECREF(first);
+	return NULL;
+    }
+    Py_DECREF(first);
+    
+    if ((gtype = pyg_type_from_object(pygtype)) == 0) {
+	return NULL;
+    }
+    if (!PyCallable_Check(callback)) {
+	PyErr_SetString(PyExc_TypeError, "third argument must be callable");
+	return NULL;
+    }
+
+    if (!g_signal_parse_name(name, gtype, &sigid, &detail, TRUE)) {
+	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     name);
+	return NULL;
+    }
+    extra_args = PySequence_GetSlice(args, 3, len);
+    if (extra_args == NULL)
+	return NULL;
+
+    data = Py_BuildValue("(ON)", callback, extra_args);
+    if (data == NULL)
+      return NULL;
+    
+    hook_id = g_signal_add_emission_hook(sigid, detail,
+					 marshal_emission_hook,
+					 data,
+					 (GDestroyNotify)pyg_destroy_notify);
+        
+    Py_DECREF(extra_args);
+    return PyInt_FromLong(hook_id);
+}
+
+static PyObject *
+pyg_remove_emission_hook(PyGObject *self, PyObject *args)
+{
+    PyObject *pygtype;
+    char *name;
+    guint signal_id;
+    gulong hook_id;
+    GType gtype;
+    
+    if (!PyArg_ParseTuple(args, "Osi:gobject.remove_emission_hook",
+			  &pygtype, &name, &hook_id))
+	return NULL;
+    
+    if ((gtype = pyg_type_from_object(pygtype)) == 0) {
+	return NULL;
+    }
+    
+    if (!g_signal_parse_name(name, gtype, &signal_id, NULL, TRUE)) {
+	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     name);
+	return NULL;
+    }
+
+    g_signal_remove_emission_hook(signal_id, hook_id);
+    
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+
 static PyMethodDef pygobject_functions[] = {
     { "type_name", pyg_type_name, METH_VARARGS },
     { "type_from_name", pyg_type_from_name, METH_VARARGS },
@@ -2258,6 +2489,9 @@ static PyMethodDef pygobject_functions[] = {
     { "markup_escape_text", (PyCFunction)pyg_markup_escape_text, METH_VARARGS|METH_KEYWORDS },
     { "get_current_time", (PyCFunction)pyg_get_current_time, METH_NOARGS },
     { "main_depth", (PyCFunction)pyg_main_depth, METH_NOARGS },
+    { "signal_accumulator_true_handled", (PyCFunction)pyg_signal_accumulator_true_handled, METH_VARARGS },
+    { "add_emission_hook", (PyCFunction)pyg_add_emission_hook, METH_VARARGS },
+    { "remove_emission_hook", (PyCFunction)pyg_remove_emission_hook, METH_VARARGS },
 
     { NULL, NULL, 0 }
 };
@@ -2448,7 +2682,9 @@ _pyg_strv_to_gvalue(GValue *value, PyObject *obj)
     int     argc, i;
     gchar **argv;
 
-    if (!PySequence_Check(obj)) return -1;
+    if (!(PyTuple_Check(obj) || PyList_Check(obj)))
+        return -1;
+
     argc = PySequence_Length(obj);
     for (i = 0; i < argc; ++i)
 	if (!PyString_Check(PySequence_Fast_GET_ITEM(obj, i)))
@@ -2457,7 +2693,6 @@ _pyg_strv_to_gvalue(GValue *value, PyObject *obj)
     for (i = 0; i < argc; ++i)
 	argv[i] = g_strdup(PyString_AsString(PySequence_Fast_GET_ITEM(obj, i)));
     argv[i] = NULL;
-    g_value_init(value, G_TYPE_STRV);
     g_value_take_boxed(value, argv);
     return 0;
 }
@@ -2929,4 +3164,7 @@ initgobject(void)
     g_log_set_handler("GThread", G_LOG_LEVEL_CRITICAL|G_LOG_LEVEL_WARNING,
                       _log_func, warning);
 
+      /* signal registration recognizes this special accumulator 'constant' */
+    _pyg_signal_accumulator_true_handled_func = \
+        PyDict_GetItemString(d, "signal_accumulator_true_handled");
 }
